// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
/*****************************************************************************
 *
 *  main.cc                                     XMM Science Analysis Subsystem
 *
 *  Main program for SAS tasks
 *
 *  (c) 1998 ESA  -  Written by Marco Beijersbergen
 *
 *  $Id: main.cc,v 1.45 2003/01/29 15:59:19 ulammers Exp $
 *
 *****************************************************************************/

#include <cstdlib>                 // C headers
#include <new>                      // install out-of-memory handler
#include <string>                   // C++ headers
#include <sstream>

#include <f90cxx.h>                 // Application headers
#include <errstr.h>
#include <Tracer.h>
#include <param.h>
#include <stringutil.h>
#include <SasFileLocator.h>
#include "HelpViewer.h"

#ifdef USE_CAL
# include <CalServer.h>
# include <Ccf.h>
#endif

#ifdef USE_OAL
# include <OalServer.h>
#endif

#ifdef USE_DAL
# include <Dal.h>
#endif

#ifdef USE_POPUP
# include "DialogDll.h"
#endif

#include "taskname.h"

namespace {
  void outOfMemoryHandler() {
    errstr<<"An out-of-memory condition has occurred: A component of the "
      "task `taskname' has tried to allocate more dynamic "
      "memory than what is currently available in the system. "
      "A likely cause of this is an attempt to read and/or write "
      "a very large data set with the DAL in its high-memory-mode. "
      "In this case, please choose the low-memory-model (see "
      "the DAL documentation for details) and try again. If this "
      "does not help it might be that too many users are using the "
      "system or the swap space is insufficient."
	  <<Sas::fatal("OutOfMemory");
  }
}


int main(int argc, char* argv[])
{
  //
  //      install out-of-memory handler
  //
  set_new_handler(outOfMemoryHandler);
  
  // Check that necessary environment variables are set
  if(getenv("SAS_DIR") == 0)
    errstr << "Environment variable SAS_DIR not defined"
	   << Sas::error("sasdir");
#ifdef NAGf90Fortran 
  f90init();
#endif
  string versionRelease("(PACKAGE-VERSION) [RELEASE]");

  // Use the taskname for error messages.
  /* hack: see also lib/Task.cc */
  if (char* envMetaTask=getenv("__SAS_META_TASK__"))
    errHandler.name(string(envMetaTask) + "::taskname");
  else
    errHandler.name("taskname");
  
  // Let the parameter interface pick up its own options
  openParameters("taskname");
  gParameters->parseCommandLine(argc,argv);
  
  bool printParameters = false;
  bool popupParameters = false;
    
#ifdef USE_CAL
  vector<string> ccfConstitReplacements;
#endif

  // Parse remaining parameters.
  int i = 1;
  while (i < argc) {
    // " -- " marks end of options
    string argument = argv[i];
    if (argument == "--") {
      argv[i] = 0;
      break;
    }
    
    string option = "", optionValue = "";
    if (argument.substr(0,2) == "--")
      option = argument.substr(2,string::npos);
    else if (argument.substr(0,1) == "-")
      option = argument.substr(1,string::npos);
    else
      option = argument;
    
    string::size_type p;
    if ((p = option.find('=')) != string::npos) {
      // option=value form
      optionValue = option.substr(p+1, option.length()-p);
      option = option.substr(0, p);
    } else {
      // a possible option value is in the next argv
      if (argv[i+1] && argv[i+1][0] != '-') {
	// this is an option value
	optionValue = argv[i+1];
	argv[i++] = 0;
      }
    }
    argv[i++] = 0;
    
    // Print help about task parameters
    if (option == "h" || option == "help") {
      cout<<"XMM-Newton Science Analysis System, task '" 
		  <<"taskname " << versionRelease <<"'\n\n"
		  <<"Usage:\n"
		  <<"  taskname [options] [--]\n\n" 
		  <<"Options:\n"
#ifdef USE_CAL
		  <<"  -a <dir1>[:<dir2>...] | --ccfpath <dir1>[:<dir2>...]\n"
#endif	
          <<"  -c | --noclobber\n"
#ifdef USE_POPUP
          <<"  -d | --dialog\n"
#endif
#ifdef USE_CAL
	      <<"  -f <f1> [<f2> ...] | --ccffiles <f1> [<f2> ...]\n"
#endif	
          <<"  -h | --help\n"
#ifdef USE_CAL
          <<"  -i <cifname> | --ccf <cifname>\n"
#endif	
          <<"  -m | --manpage\n"
#ifdef USE_OAL
          <<"  -o <odfname> | --odf <odfname>\n"
#endif	
          <<"  -p | --param\n"
          <<"  -t | --trace\n"
          <<"  -V <level> | --verbosity <level>\n"
          <<"  -v | --version\n"
          <<"  -w [code|n] | --warning [code|n]\n"
          <<"  --<param>=<value>\n"
		  <<"  --\n\n"
		  <<"Parameters:"
		  <<endl;
      gParameters->writeHelpToStream(cout);
      return 0;
    }
    
    else if (option == "p" || option == "param") {
      printParameters = true;
      continue;
    }
    
    else if (option == "t" || option == "trace")
      tracer.trace(true);
    
    /*
    // Specify output file
    else if (option == "o") {
      if (optionValue.length() == 0)
	errstr << "missing argument for -o"
	       << Sas::error("missingOutputFile");
      setParameter("output", optionValue);
    }
    */

    // Print version number and exit
    else if (option == "v" || option == "version") {
      cout << "taskname" << ' ' << versionRelease << endl;
      exit(0);
    }
    
    // HTML help
    else if (option == "m" || option == "manpage") {
      string url = "file://" + fileLocator.locate("doc/taskname/index.html", "Cannot find documentation for taskname.");
      if(url.size() == 0) exit(1);
      
      if (! HelpViewer().browse(url)) exit(1);
      exit(0);
    }
    
    // Set message verbosity level
    else if (option == "V" || option == "verbosity") {
      if (optionValue.length() == 0)
	errstr << "missing argument for verbosity"
	       << Sas::error("missingVerbosity");
      stringstream ss;
      ss << optionValue;
      int level;
      ss >> level;
      if (!ss)
	errstr << "invalid verbosity level " << optionValue << Sas::error("verbosityLevel");
      errHandler.verbosity(level);
    }
    // Set message verbosity level
    else if (option == "w" || option == "warning") {
      errHandler.ignoreWarning(optionValue);
    }
    
#ifdef USE_CAL
    //
    //  CAL specific options
    //
    else if (option == "i" || option == "ccf") {
      if (optionValue.empty())
        errstr<<"missing argument for ccf"<<Sas::error("missingCcf");
      
      calServer.setCcf(optionValue);
    }

	else if (option == "a" || option == "ccfpath") {
      if (optionValue.empty())
        errstr<<"missing argument for ccfpath"<< Sas::error("missingCcfPath");
      calServer.ccf()->ccfSearchPath(optionValue);
    }

    else if (option == "f" || option == "ccffiles") {
      if (optionValue.length() == 0)
       errstr<<"missing argument for ccffiles"<< Sas::error("missingCcfFiles");

      ccfConstitReplacements = breakup(optionValue);
    }
#endif
    
#ifdef USE_OAL
    //
    //  OAL specific options
    //
    else if (option == "o" || option == "odf") {
      if (optionValue.length() == 0)
	errstr << "missing argument for odf"
	       << Sas::error("missingOdfdir");
      
      oalServer.newOdf(optionValue);
    }
#endif
    
#ifdef USE_DAL
    else if (option == "c" || option == "noclobber")
      dataSetServer->clobber(false);
#endif	
    
#ifdef USE_POPUP
    else if (option == "d" || option == "dialog")
      popupParameters=true;
#endif

    else errstr << "invalid option '" << argument << "'"
		<< Sas::error("invalidOption");
    
  }
  
  // Remove all arguments that have been set to 0
  for (int i = 0; i < argc; ) {
    if (argv[i]) ++i;
    else {
      for (int j = i; j < argc-1; ++j)
	argv[j] = argv[j+1];
      --argc;
    }
  }
  
#ifdef USE_CAL
	//	retarded registration of new CCF constituents to allow the
	//	CCF search path (--ccfpath) to be set
	if (ccfConstitReplacements.size()) {
		Ccf *ccf = calServer.ccf();
		for (unsigned i=0; i<ccfConstitReplacements.size(); ++i)
			ccf->registerNewCifEntry(ccfConstitReplacements[i]);
		//
		//	restart atom server to force flushing of memory buffered atoms
		//
		calServer.atomServer()->restart();
	}
#endif


  /*
  // The remaining argument are transformed into parameters infile if only one
  // and infile1 .. infileN if there are multiple.
  switch (argc) {
  case 0: break;
  case 1: break;
  case 2: setParameter("input",argv[1]); break;
  default: 
    {
      for (int i = 1; i < argc; ++i) {
	stringstream s;
	s << "input" << i;
	setParameter(s.str(),argv[i]);
      }
    }
  }
  */

  if (popupParameters) {
# ifdef USE_POPUP
    DialogDll popup;
    if (!popup.exec(gParameters)) {
      // what do we want to do then?
      return 1;
    }
# endif
  }

  // Print the parameters
  if (printParameters) {
    cout << *gParameters;
    gParameters->checkMandatoryParameters();
    return 0;
  }
  
  // Check that all mandatory parameters have been specified.
  // We do this AFTER processing other options, so that '-v', '-h', etc
  // can still be used without supplying mandatory parameters.
  if (!popupParameters) gParameters->checkMandatoryParameters();
    
  
  task(taskname);

#ifdef USE_CAL
  errstr<<"List of CCF constituents accessed during this run:\n"
		<<calServer.ccfAccessLog()<<endl
		<<Sas::message(Msg::App, Msg::Verbose);
#endif


  /* f90/f95 bug: sdterr is closed when using: f90finish().
     f90finish();
  */
    
  closeParameters();
  
  return 0;
}
